"""

File:    PbxConfigJobs.py
Author:  Jonathan Maurice <jmaurice@sangoma.com>
Date:    4/6/2011 5:58:04 PM
 
Description: Contain jobs which are run by the PbxConfigJobRunner. Those jobs
are not Converter therefore they are implemented here.
 
Copyright (C) 2003-2011 Sangoma Technologies. 
All Rights Reserved.

This code is Sangoma Technologies Confidential Property.
Use of and Access to this code is covered by a previously executed
Non-disclosure Agreement between Sangoma Technologies and the Recipient.
This code is being supplied for evaluation purposes only and is not to be
used for any other purpose.  

"""

import os

from paraxip.infra.PyLoggingFunctions import  *
from netborderPyInfra import TraceScope
from paraxip.infra.PyJob import Job
from paraxip.smart_update.ExportCfg import ConfigExporter
import paraxip.smart_update.ImportCfg as ImportCfg

fsLogger = Logger("netborder.pbxConfigurator.PbxConfigJob")
smartUpJobLogger = \
  Logger("netborder.pbxConfigurator.PbxConfigJob.SmartUpdateJob")


class PbxConfigJob(Job):
  """PbxConfigJob interface"""

  def __init__(self, jobStatus, pbxConfigManager):
    t = TraceScope(fsLogger, "PbxConfigJob.__init__")
    super(PbxConfigJob, self).__init__()
    self.jobStatus = jobStatus
    self.pbxConfigManager = pbxConfigManager

  def run(self):
    t = TraceScope(fsLogger, "PbxConfigJob.run")
    raise NotImplementedError("PbxConfigJob.run should not be called." \
                              "Abstract class!")
    return False


class SmartUpdateJob(PbxConfigJob):
  """Export the created configuration and then import it to update the
  configuration to the current version of the application."""

  def __init__(self, jobStatus, pbxConfigManager):
    t = TraceScope(smartUpJobLogger, "SmartUpdateJob.__init__")
    super(SmartUpdateJob, self).__init__(jobStatus, pbxConfigManager)
    
  def run(self):
    """Do the smart update process on the generated config files"""
    t = TraceScope(smartUpJobLogger, "PbxConfigJobs.run")

    #Export configuration in srcDir
    configDirPath = self.pbxConfigManager.getSrcDir()
    binDirPath = self.pbxConfigManager.getBinDirectory()
    gwExePath = os.path.join(binDirPath, 'netborder-gateway')
    #XXX_INSTDIR_XXX will be replace by the real installation directory.
    gwInstPath = 'XXX_INSTDIR_XXX'
    gwVersion = self.pbxConfigManager.getAppVersion()

    exportArchiveName = \
        os.path.splitext(self.pbxConfigManager.getPbxFileName())[0]
    exportArchiveName = exportArchiveName + '.zip'
    exportDirPath = self.pbxConfigManager.getWorkingDirectory()
    exportArchivePath = os.path.join(exportDirPath, exportArchiveName)
    
    cfgExporter = ConfigExporter(configDirPath, exportDirPath,
                                 gwExePath, gwInstPath,
                                 exportFileName='export.txt',
                                 archiveName=exportArchiveName)

    #Set the gateway version to be the same as in the pbx file.
    cfgExporter.setAppInfo(strAppVersion=gwVersion)

    log_debug(smartUpJobLogger, "Exporting configuration %s to %s" %
             (configDirPath, exportArchivePath)
    )
    if cfgExporter.export() is not True:
      log_error(smartUpJobLogger, 
                "Failed to export configuration at %s." % configDirPath
      )
      return False

    #Import the configuration.
    outputDirPath = self.pbxConfigManager.getDstDir()

    log_debug(smartUpJobLogger, "Importing configuration %s to %s" %
             (exportArchivePath, outputDirPath)
    )
    rc = ImportCfg.importCfg(exportArchivePath, outputDirPath)
    if rc != 0:
      log_error(smartUpJobLogger, 
                "Failed to import configuration at %s. Cause=%s" %
                (exportArchivePath, rc)
      )
      return False

    log_debug(smartUpJobLogger, "Update of configuration file at %s to %s "\
              "succeeded"% (configDirPath, outputDirPath)
    )
    return True

###############################################################################
# Runtest
if __name__ == '__main__':

  import shutil
  from paraxip.pbx_configurator.PbxConfigManager import PbxConfigManager
  from paraxip.infra.PyJob import JobStatus
  from paraxip.pbx_configurator.PbxConfigurationInfo import PbxConfigurationInfo
  
  testFail = False

  Logger('netborder.smartUpdate').setLogLevel(LogLevel.debug)
  fsLogger.setLogLevel(LogLevel.debug)

  pbxFile = os.path.join('test-input', '3cx-a200-2fxo-2fxs.pbx')
  cfgDir = os.path.join('test-output', 'smartUpdateJob-test')
  replaceStrDict = {'na_me':'name'}

  pbxCfgInfo = PbxConfigurationInfo(pbxFilePath=pbxFile)

  testStatus = JobStatus()
  cfgMan = PbxConfigManager(pbxFile, pbxCfgInfo, cfgDir,
                            replaceStrDict, '.pbx')

  if not cfgMan.initialize():
    log_error(fsLogger, "Failed to initialize PbxConfigManager")
    testFail = True

  if not cfgMan.initialConvert():
    log_error(fsLogger, "PbxConfigManager initial convert failed")
    testFail = True

  if not cfgMan.setTempMode():
    log_error(fsLogger, "Failed to set PbxConfigManager to temp mode")
    testFail = True
  
  #Copy file from test-input/smart-update-job-test to srcDir
  cfgToUpdateDir = os.path.join('test-input', 'smart-update-job-test')
  srcDir = cfgMan.getSrcDir()

  fileToCopy = ['gw.properties', 'pstn-config.xml',
                'routing-rules.xml', 'sip-client-registration.xml']

  for file in fileToCopy:
    try:
      srcFilePath = os.path.join(cfgToUpdateDir, file)
      dstFilePath = os.path.join(srcDir, file)
      shutil.copy2(srcFilePath, dstFilePath)
    except (OSError, IOError), e:
      log_debug(fsLogger, "Failed to copy file %s to %s" % 
                (srcFilePath, dstFilePath))

  #Update the config that was copied.
  smartUpJob = SmartUpdateJob(testStatus, cfgMan)
  if not smartUpJob.run():
    log_error(fsLogger, "Failed to update configuration at %s" %
             cfgMan.getSrcDir())
    testFail = True

  if testFail:
    log_error(fsLogger, "Runtest failed!")
    print("Failed")
  else:
    log_info(fsLogger, "Runtest passed!")
    print("Passed")
## !!MIRROR FILE, DO NOT MODIFY!!
