#!/bin/sh

###############################################################################
# Title:  configure-snmpd.sh
# Author: Jean-Philippe Boily, jr. eng.
# Date:   2011-01-17
# Requirements: 
#   - Prior to using this script, you need to make sure that Net-SNMP package 
#     is installed on the computer. This can be accomplished using Yum or any 
#     other package manager. 
#   - The configuration file for snmpd process MUST be located in 
#     /etc/snmp/snmpd.conf.
# Description:
#   The goal of this script is to add the configuration lines required by the
#   SNMP Server agent to accept SNMP subagent connection requests. If the
#   configuration file /etc/snmp/snmpd.conf exists, the script will start by
#   searching in the file if the necessary configuration lines are already
#   there. Those lines are:
#     - rocommunity public: the gateway snmp subagent uses public community.
#     - master agentx: sets the snmpd process as a endpoint where subagents
#                      will connect.
#     - agentxsocket xyz: The address where subagents will send their 
#                         connection requests. Default value for xyz is 
#                         "tcp:localhost:705".
#   For each of those lines, if one is missing, we append it to the end of the
#   configuration file. If the agentxsocket line exists, we extract its xyz
#   value and assign it as the value for the parameter "snmp.agentx.IPAddress"
#   in the gateway configuration file $INSTALL_DIR/config/gw.properties.
#
#   It is also possible that snmpd's configuration file defines views that are
#   used to specify, for different group of users, the subtrees they are
#   allowed to read and/or write to. After a fresh install of Net-SNMP, those
#   views will be defined like this:
#
#             name           incl/excl     subtree         mask(optional)
#     view    systemview    included   .1.3.6.1.2.1.1
#     view    systemview    included   .1.3.6.1.2.1.25.1.1
#
#   To expose all the gateway SNMP fields, two subtrees MUST be availble to
#   the SNMP clients: .1.3.6.1.2.1.10 and .1.3.6.1.4.1.35987. Since most of
#   the time the default configuration still applies, the two following 
#   entries will be added to the snmpd.conf file by this script:
#
#     view    systemview    included   .1.3.6.1.2.1.10
#     view    systemview    included   .1.3.6.1.4.1.35987
#   
#   The OID 35987 is the registered Sangoma Technologies Corp. OID.
#
###############################################################################

SNMPD_CONFIG_FILE=/etc/snmp/snmpd.conf
GW_PROPERTIES_FILE=XXX_INSTDIR_XXX/config/gw.properties
ACCESS_CONTROL_STR="rocommunity public"
OPERATION_MODE_STR="master agentx"
AGENTX_SOCKET_STR="agentxsocket"
DEFAULT_AGENTX_ADDR_STR="tcp:localhost:705"
DS0_DS1_ROOT_OID_STR=".1.3.6.1.2.1.10"
SANGOMA_PRIVATE_OID_STR=".1.3.6.1.4.1.35987"
DEFAULT_VIEW_STR="view    systemview    included"

# varialbe that defines if snmpd service needs to be restarted
bRestartSnmpd="false"

# Try to find the default snmpd config file
if [[ -f $SNMPD_CONFIG_FILE ]] ; then
  configLinesHeader=$(grep -m1 "Netborder Express Gateway - SNMP" < $SNMPD_CONFIG_FILE)
  if [[ ${#configLinesHeader} == 0 ]] ; then
    # Add NBE information to the snmpd.conf file
    echo "
################################################################################
#
# Netborder Express Gateway - SNMP configuration requirements
#
# NBE Gateway has a builtin SNMP subagent that requires a SNMP Server agent
# with specific configuration lines. The following configuration lines or an
# equivalent must be present in this file for the NBE Gateway subagent to be
# able to connect to the agent.
#
#   - rocommunity public: the public community must be defined.
#   - master agentx: allow the SNMP agent to receive connection requests from
#                   subagents using agentx protocol.
#   - agentxsocket xyz: defines the agentx socket address where connection
#                       requests will be sent by the subagent. The default value
#                       for this field is tcp:localhost:705.
#
# If this configuration file uses views to define, for different group of users,
# the subtrees they are allowed to read and/or write to, two subtrees MUST be 
# availble to the SNMP clients: .1.3.6.1.2.1.10 and .1.3.6.1.4.1.35987. Those
# OID are the roots of the fields exposed by Netborder Express Gateway. Since 
# most of the time the default configuration still applies, the two following
# entries will be added to the file:
#
#   view    systemview    included    .1.3.6.1.2.1.10
#   view    systemview    included    .1.3.6.1.4.1.35987
#
# The OID 35987 is the registered Sangoma Technologies Corporation OID. If those
# views does not match you configuration, please add views for those OID 
# manually.
#
################################################################################
    " >> $SNMPD_CONFIG_FILE
  fi

  # look for agentx entries
  accessControlStr=$(grep -m1 -P "^$ACCESS_CONTROL_STR" < $SNMPD_CONFIG_FILE)
  operationModeStr=$(grep -m1 -P "^$OPERATION_MODE_STR" < $SNMPD_CONFIG_FILE)
  agentxSocketStr=$(grep -m1 -P "^$AGENTX_SOCKET_STR" < $SNMPD_CONFIG_FILE)
  ds0ds1OIDStr=$(grep -m1 -P "^$DEFAULT_VIEW_STR    $DS0_DS1_ROOT_OID_STR" < $SNMPD_CONFIG_FILE)
  sangomaOIDStr=$(grep -m1 -P "^$DEFAULT_VIEW_STR    $SANGOMA_PRIVATE_OID_STR" < $SNMPD_CONFIG_FILE)

  # Append access control string
  if [[ ${#accessControlStr} == 0 ]] ; then
    echo "$ACCESS_CONTROL_STR" >> $SNMPD_CONFIG_FILE
    bRestartSnmpd="true"
  fi

  # Append operation mode string
  if [[ ${#operationModeStr} == 0 ]] ; then
    echo "$OPERATION_MODE_STR" >> $SNMPD_CONFIG_FILE
    bRestartSnmpd="true"
  fi

  # Append agentxsocket string
  if [[ ${#agentxSocketStr} == 0 ]] ; then
    echo "$AGENTX_SOCKET_STR $DEFAULT_AGENTX_ADDR_STR" >> $SNMPD_CONFIG_FILE
    bRestartSnmpd="true"
  else
    # Make sure the gateway properties file exists
    if [[ -f $GW_PROPERTIES_FILE ]] ; then
      # Extract current agentxsocket address
      socketAddr=$(echo $agentxSocketStr | sed 's|agentxsocket ||')
      sed -si "s|snmp.agentx.IPAddress=.*|snmp.agentx.IPAddress=${socketAddr}|" $GW_PROPERTIES_FILE
    else
      echo "ERROR: the gateway properties file $GW_PROPERTIES_FILE does not exists"
    fi
  fi

  # Add default views to the config file
  if [[ ${#ds0ds1OIDStr} == 0 ]] ; then
    echo "$DEFAULT_VIEW_STR    $DS0_DS1_ROOT_OID_STR" >> $SNMPD_CONFIG_FILE
    bRestartSnmpd="true"
  fi

  if [[ ${#sangomaOIDStr} == 0 ]] ; then
    echo "$DEFAULT_VIEW_STR    $SANGOMA_PRIVATE_OID_STR" >> $SNMPD_CONFIG_FILE
    bRestartSnmpd="true"
  fi

  if [[ "$bRestartSnmpd" == "true" ]] ; then
    echo "Restarting snmpd service"
    service snmpd restart
  fi
else
  echo "ERROR: failed to install service. Missing configuration file $SNMPD_CONFIG_FILE"
  exit 1
fi

