import logging

import os
import stat
import simplejson
from netborderPyInfra import PyGlobalConfig as globalConfig
from turbogears import controllers, expose, error_handler, validate
from turbogears import flash, redirect, url
from cherrypy.lib.cptools import serve_file

log = logging.getLogger("pstnconfig")

class FileBrowserController(controllers.Controller):
  installDir = globalConfig.getParameter("netborder.Installation.Directory")
  log.info("FileBrowserController::installDir var is: %s" % installDir)

  # This array specifies the file extension we want to be
  # available in the file browser. The utility is to prevent
  # .exe, .dll, etc. files to be displayed.
  allowedFileExtension = [
    "txt",
    "out",
    "log",
    "properties",
    "cfg",
    "xml",
    "conf",
    "sh",
    "wav",
    "pdf",
    "zip"
  ]

  # In this array, we specify complete file name with the extension
  # ("filename.ext") we absolutely don't want to be displayed even
  # if the extension would allowed the file to be shown to the user.
  notShownFiles = []

  # In this array, we put the files ("filename.ext") we want the user
  # to be able to read in the browser, but not be able to modify
  # their content.
  readOnlyFiles = [
    "GatewayDebug.log",
    "GatewayRollingFile.log",
    "OAMWebServerDebug.log",
    "OAMWebServerRollingFile.log",
    "TurboGears.debug.log",
    "License.txt",
    "netborder-express-gateway-license.txt",
    "netborder-gateway-license.txt",
    "T_gw.properties",
    "T_routing-rules.xml"
  ]

  # This array contains all the directories to be hidden to the user
  notShownDirectory = [
    ".svn",
    "tones",
    "Stubs",
    "Plugins",
    "web",
    "drivers",
    "oamWebServer",
    "Python-2.5"
  ]

  @expose(template='pstnconfig.templates.fileBrowser')
  def lightbox_file_browser(self):
    return {"footerText":""}

  #######################################################
  # This is the entry method of the file browser.
  # The arguments are as followed:
  #   cmd: get, download
  #   node: id of the node in the tree
  #   path: relative path to the file 
  # - Get command builds a JSON string to be represented
  # in the fileTreePanel.
  # - Download makes the browser to download the file.
  #######################################################
  @expose()
  def index (self, cmd, node='', path=''):
    # The paths are relative to root keyword to hide the absolute
    # path to the user. Replace root with installation directory.
    path = path.replace("root",self.installDir)

    if cmd == 'get':
      data = self.getDirectoryFilesDict(path)
      # The fileTreePanel needs a JSON string representation
      # of the directory tree.
      return simplejson.dumps(data)

    elif cmd == 'download':
      # Use cherryPy method to serve the file in the browser
      # as an attachment which cause the browser to open a
      # download window.
      return serve_file(path, "application/x-download", "attachment")

  #######################################################
  # Builds an array of dictionaries that contains the 
  # informations on the file to be displayed.
  # @params dirPath: path to the directory wich need to 
  #                 be listed.
  #######################################################
  def getDirectoryFilesDict(self, dirPath):
    data = []
    rootDirs = os.listdir(dirPath)
    for dirFile in rootDirs:
      (filename,extension) = os.path.splitext(dirFile)
      if extension:
        extension = extension.split(".")[1]
       
      if os.path.isdir(os.path.join(dirPath, dirFile)) \
        and ( dirFile not in self.notShownDirectory ):
        data.append({
          'text':dirFile,
          'disabled':False,
          'leaf':False
        })

      elif os.path.isfile(os.path.join(dirPath, dirFile)) \
        and (dirFile not in self.notShownFiles):

        backupFile = False
        if not extension:
          readOnly = True
        elif extension.startswith("bak") or extension == "log":
          readOnly = True
          backupFile = True
        else:
          readOnly = (dirFile in self.readOnlyFiles)

        if (not extension) or \
          (extension in self.allowedFileExtension) or \
          backupFile :
          data.append({
            'text':dirFile,
            'iconCls':"file-%s" % extension,
            'disabled':readOnly,
            'leaf':True,
            'extension':extension
          })

    return data

  #######################################################
  # Get the size of the file.
  # @params filePath: path to the file to read.
  #######################################################
  @expose()
  def getFileSize(self, filePath):
    filePath = filePath.replace("/root",self.installDir)
    
    f = os.stat(filePath)
    size = f[stat.ST_SIZE]

    return {'fileSize':size}

  #######################################################
  # Get the number items in the directory.
  # @params filePath: path to the directory.
  #######################################################
  @expose()
  def getNumItemsInDir(self, filePath):
    filePath = filePath.replace("/root",self.installDir)

    return {"numOfItems":len(os.listdir(filePath))}

  #######################################################
  # Try to open the file and return its content to the 
  # browser so it can be displayed in the text area.
  # @params filePath: path to the file to read.
  #######################################################
  @expose()
  def showFileInBrowser(self, filePath):
    log.info("Show File in browser: %s" % filePath)

    filePath = filePath.replace("/root",self.installDir)
    
    try:
      f = open(filePath,'r')
      # Read all content
      strFileContent = f.read()
      f.close()
    except IOError:
      log.error("Cannot open file %s" % filePath)               
      return ""
    return {'fileText':strFileContent}

  #######################################################
  # Saves the new file content.
  # @params filePath: path to the file to save.
  # @params fileContent: the new content of the file.
  #######################################################
  @expose()
  def saveFileChanges(self, filePath, fileContent):
    log.info("Saving File : %s" % filePath)

    filePath = filePath.replace("/root",self.installDir)
    
    try:
      f = open(filePath,'wb')
      # Write new content
      f.write(fileContent)
      f.close()
    except IOError:
      log.error("Cannot open file %s" % filePath)               
      return {}

    return {}

